/*
 * %CopyrightBegin%
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Copyright Ericsson AB 2015-2025. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * %CopyrightEnd%
 */

package com.ericsson.otp.erlang;

import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;

/**
 * Default socket-based server transport
 * 
 * @author Dmitriy Kargapolov
 */
public class OtpServerSocketTransport implements OtpServerTransport {

    /**
     * Underlying server socket
     */
    private final ServerSocket socket;

    /**
     * @see ServerSocket#ServerSocket(int)
     */
    public OtpServerSocketTransport(final int port) throws IOException {
        socket = new ServerSocket(port);
    }

    /**
     * @see ServerSocket#getLocalPort()
     */
    public int getLocalPort() {
        return socket.getLocalPort();
    }

    /**
     * @see ServerSocket#accept()
     */
    public OtpTransport accept() throws IOException {
        final Socket sock = socket.accept();
        sock.setTcpNoDelay(true);
        return new OtpSocketTransport(sock);
    }

    /**
     * @see ServerSocket#close()
     */
    public void close() throws IOException {
        socket.close();
    }

}
